<?php 
/**
 * Model taula 'incidencia' i relacionades
 * Basat en MY_Model (funcionalitat CRUD)
 */

if (!defined('BASEPATH')) exit('Accés denegat');

class Incidencia_model extends MY_Model
{

    public $_table = 'incidencia';
    public $primary_key = 'id_incidencia';
    public $before_create = array( '_create_timestamp' );
    public $before_update = array( '_update_timestamp' );
    
    public function __construct()
    {
        parent::__construct();
    }
    
    /**
     * Recupera dades incidència
     * @param {string} $id_incidencia
     * @return {object} fila
     */
    public function get_incidencia($id_incidencia)
    {
        $this->db->select('incidencia.*, motiu.descripcio motiu_descripcio, estat.nom estat_nom,
            concat(solicitant.nom," ",solicitant.cognoms) solicitant_nom, (CASE WHEN solicitant.foto IS NULL THEN "generica.png" ELSE solicitant.foto END) as solicitant_foto,
            concat(gestor.nom," ",gestor.cognoms) gestor_nom, (CASE WHEN gestor.foto IS NULL THEN "generica.png" ELSE gestor.foto END) as gestor_foto,
            departament.nom departament', FALSE);
        $this->db->from('incidencia');
        $this->db->join('estat', 'estat.id_estat = incidencia.estat');
        $this->db->join('usuari gestor', 'gestor.id_usuari = incidencia.gestor_assignat', 'left');
        $this->db->join('usuari solicitant', 'solicitant.id_usuari = incidencia.solicitant', 'left');
        $this->db->join('departament', 'departament.id_departament = incidencia.departament_assignat', 'left');
        $this->db->join('motiu', 'motiu.id_motiu = incidencia.motiu');
        $this->db->where('id_incidencia', $id_incidencia);
        $res = $this->db->get();
        return $res->row();
    }
    
    /**
     * Obté incidències segons filtres donats
     * @param {array} $filtres Array de filtres
     * @param {string} $limit Màxim d'incidències a retornar
     * @param {string} $inici Punter inici pel qual començar a retornar incidències
     * @return {array} Incidències
     */
    public function get_incidencies($filtres = array(), $limit = '0', $inici = '0')
    {
        $this->db->select('incidencia.*, estat.nom estat_nom, concat(gestor.nom," ",gestor.cognoms) gestor_nom, gestor.foto gestor_foto, departament.nom departament_nom', FALSE);
        $this->db->from('incidencia');
        $this->db->join('estat', 'estat.id_estat = incidencia.estat');
        $this->db->join('usuari gestor', 'gestor.id_usuari = incidencia.gestor_assignat', 'left');
        $this->db->join('departament', 'departament.id_departament = incidencia.departament_assignat', 'left');
        if (@$filtres['text'])
        {
            $text = $this->db->escape_like_str($filtres['text']);
            $this->db->where('(titol LIKE "%'.$text.'%" OR incidencia.descripcio LIKE "%'.$text.'%")', NULL, TRUE);
        }
        if (count(@$filtres['estats'])>0)
            $this->db->where_in('estat', $filtres['estats']);
        if (@$filtres['prioritat'])
            $this->db->where('prioritat', $filtres['prioritat']);
        if (@$filtres['motiu'])
            $this->db->where('motiu', $filtres['motiu']);
        if (@$filtres['desde'])
            $this->db->where('incidencia.added_on >=', $filtres['desde'], TRUE);
        if (@$filtres['fins'])
            $this->db->where('incidencia.added_on <=', 'DATE_ADD("'.$this->db->escape_like_str($filtres['fins']).'", INTERVAL 1 DAY)', FALSE);
        if (@$filtres['solicitant'])
            $this->db->where('solicitant', $filtres['solicitant']);
        if (@$filtres['gestor'])
            $this->db->where('gestor_assignat', $filtres['gestor']);
        if (@$filtres['departament'])
            $this->db->where('departament_assignat', $filtres['departament']);
        if (@$filtres['upd_desde'])
            $this->db->where('incidencia.updated_on >=', 'DATE_ADD(NOW(), INTERVAL -'.$filtres['upd_desde'].' DAY)', FALSE);

        $this->db->order_by('added_on', 'desc');
        
        if ($limit!='0' || $inici!='0')
            $this->db->limit($limit, $inici);
        
        $res = $this->db->get();
        return $res->result();
    }
    
    /**
     * Guarda incidència amb les dades enviades per formulari
     * @param {Array} $post Dades enviades per formulari
     * @param {String} $adjunt Nom del fitxer adjuntat, si existeix
     * @return {int} ID de la nova incidència
     */
    public function insert_incidencia($post, $adjunt)
    {
        $motiu = $this->motiu_model->get($post['motiu']);
        
        $data = array(
            'solicitant' => $this->session->userdata('usuari')->id_usuari,
            'motiu' => $post['motiu'],
            'titol' => $post['titol'],
            'descripcio' => nl2br($post['descripcio']),
            'estat' => '1',
            'prioritat' => $post['prioritat'],
            'departament_assignat' => $motiu->id_departament,
            'fitxer_adjunt' => $adjunt
        );
        return $this->incidencia_model->insert($data);
    }
    
    /**
     * Actualitza l'estat de la incidència amb l'últim abans de l'actual
     * @param {string} $incidencia ID incidència
     * @return {string} Nou estat assignat
     */
    public function upd_ultim_estat($incidencia, $estat_actual)
    {
        $nou_estat = $this->_log_get_ultim_estat($incidencia);
        $this->incidencia_model->update($incidencia, array('estat' => $nou_estat));
        $this->_log_set_nou_estat($incidencia, $nou_estat, $estat_actual);
        return $nou_estat;
    }
    
    /**
     * Retorna últim estat d'una incidencia
     * @param {string} $incidencia ID incidència
     * @param {string} $quan Per defecte 'abans' de l'actual, però si s'especifica 'despres' retorna estat actual
     * @return {string} estat
     */
    protected function _log_get_ultim_estat($incidencia, $quan = 'abans')
    {
        $this->db->select($quan);
        $this->db->order_by('added_on', 'desc');
        $q = $this->db->get_where('actualitzacio', array('incidencia' => $incidencia, 'tipus' => 'estat'), '1');
        return $q->row()->$quan;
    }
    
    /**
     * Guarda actualització estat incidència
     * @param {string} $incidencia ID incidència
     * @param {string} $nou_estat ID nou estat
     * @param {string} $antic_estat ID antic estat
     */
    protected function _log_set_nou_estat($incidencia, $nou_estat, $antic_estat = "")
    {
        $antic_estat = (empty($antic_estat)) ? $this->_log_get_ultim_estat($incidencia, 'despres') : $antic_estat;
        
        $data = array(
            'incidencia' => $incidencia,
            'tipus' => 'estat',
            'abans' => $antic_estat,
            'despres' => $nou_estat,
            'added_by' => $this->session->userdata('usuari')->id_usuari
        );
        $this->db->insert('actualitzacio', $data);
    }
    
    /**
     * Actualitza l'estat d'una incidència i registra l'actualització
     * @param {string} $incidencia
     * @param {string} $estat_antic
     * @param {string} $estat
     */
    public function upd_estat($incidencia, $estat_antic, $estat)
    {
        $this->db->where('id_incidencia', $incidencia);
        $this->db->update('incidencia', $this->_update_timestamp(array('estat' => $estat)));
        $this->_log_set_nou_estat($incidencia, $estat, $estat_antic);
    }
    
    /**
     * Actualitza el gestor, el departament assignats i l'estat d'una incidència
     * i registra l'actualització
     * @param {string} $incidencia
     * @param {string} $departament
     * @param {string} $gestor
     * @param {string} $estat
     * @return {string} Estat resultant de la incidència
     */
    public function upd_assignacions($incidencia, $assignacio_antiga, $departament, $departament_nom, $gestor, $estat)
    {
        $data = array();
        if ($gestor!='')
            $data['gestor_assignat'] = $gestor;
        else
            $data['gestor_assignat'] = NULL;
        if ($estat=='1' && $gestor!='')
        {
            $data['estat'] = $estat = '2';
        }
        elseif ($estat=='2' && $gestor=='')
        {
            $data['estat'] = $estat = '1';
        }
        $data['departament_assignat'] = $departament;
        $this->db->where('id_incidencia', $incidencia);
        $this->db->update('incidencia', $this->_update_timestamp($data));
                
        $data = array(
            'incidencia' => $incidencia,
            'tipus' => 'assignacio',
            'abans' => $assignacio_antiga,
            'despres' => $departament_nom.' / '.$gestor,
            'added_by' => $this->session->userdata('usuari')->id_usuari
        );
        $this->db->insert('actualitzacio', $data);
        
        return $estat;
    }
    
    /**
     * Actualitza la prioritat assignada d'una incidència i registra l'actualització
     * @param {string} $incidencia
     * @param {string} $prioritat_antiga
     * @param {string} $prioritat
     */
    public function upd_prioritat($incidencia, $prioritat_antiga, $prioritat)
    {
        $this->db->where('id_incidencia', $incidencia);
        $this->db->update('incidencia', $this->_update_timestamp(array('prioritat' => $prioritat)));
        
        $data = array(
            'incidencia' => $incidencia,
            'tipus' => 'prioritat',
            'abans' => $prioritat_antiga,
            'despres' => $prioritat,
            'added_by' => $this->session->userdata('usuari')->id_usuari
        );
        $this->db->insert('actualitzacio', $data);
    }
    
    /**
     * Obté totes les actualitzacions d'una incidència donada
     * @param {string} $incidencia ID Incidència
     * @return {array} Actualitzacions (o FALSE si no n'hi ha cap)
     */
    public function get_actualitzacions($incidencia)
    {
        $this->db->select('DATE_FORMAT(actualitzacio.added_on, "%d/%m/%Y %H:%i") data,
                CONCAT(usuari.nom, " ", usuari.cognoms, " (",actualitzacio.added_by,")") usuari_nom,
                CONCAT(UCASE(SUBSTRING(actualitzacio.tipus, 1, 1)),LCASE(SUBSTRING(actualitzacio.tipus, 2))) tipus,
                CONCAT(
                    "De \"",
                    (CASE WHEN abans IS NULL THEN \'-\' WHEN estat_abans.nom IS NULL THEN abans ELSE estat_abans.nom END),
                    "\" a \"",
                    (CASE WHEN estat_despres.nom IS NULL THEN despres ELSE estat_despres.nom END),"\"") canvi', FALSE);
        $this->db->from('actualitzacio');
        $this->db->join('estat estat_abans', 'estat_abans.id_estat = abans', 'left');
        $this->db->join('estat estat_despres', 'estat_despres.id_estat = despres', 'left');
        $this->db->join('usuari', 'usuari.id_usuari = actualitzacio.added_by');
        $this->db->where('incidencia', $incidencia);
        $this->db->order_by('actualitzacio.added_on', 'desc');
        $q = $this->db->get();
        if ($q->num_rows() > 0) {
            $data = array();
            foreach ($q->result_array() as $fila) { // ->result retorna objecte
                $data[] = $fila;
            }
            return $data;
        }
        return FALSE;
    }
    
    /**
     * Obté el número d'incidències agrupades per motiu obertura
     * @param type $dies Interval dies a tenir en compte
     * @return {array} Motius i nº d'incidències
     */
    public function get_top_motius($dies)
    {
        $this->db->select('motiu.descripcio as motiu, count(incidencia.id_incidencia) as incidencies');
        $this->db->from('incidencia');
        $this->db->join('motiu', 'motiu.id_motiu = incidencia.motiu');
        $this->db->where('incidencia.updated_on>=', 'DATE_ADD(NOW(), INTERVAL -'.$dies.' DAY)', FALSE);
        $this->db->group_by('incidencia.motiu');
        return $this->db->get()->result_array();
    }
    
    /**
     * Obté 10 gestors amb més incidències obertes
     * @return {object} Resultat
     */
    public function get_carrega_gestors()
    {
        $q = $this->db->query("SELECT (CASE WHEN gestor_assignat IS NULL THEN 'No assignades' ELSE CONCAT(usuari.nom,' ',usuari.cognoms) END) gestor, count(id_incidencia) obertes
                            FROM incidencia
                            LEFT JOIN usuari ON usuari.id_usuari = gestor_assignat
                            WHERE incidencia.estat IN ('1','2','3')
                            GROUP BY gestor_assignat
                            ORDER BY obertes DESC
                            LIMIT 10");
        return $q->result();
    }
    
    /**
     * Retorna activitat (noves incidències i actualitzacions) durant el període desitjat.
     * @param {int} $periode Període en segons
     * @return {array}
     */
    public function get_activitat_periode($periode = 10)
    {
        $q = $this->db->query('SELECT sum(activitat) activitat
            FROM (
                SELECT count(*) activitat FROM incidencia WHERE updated_on >= DATE_ADD(NOW(), INTERVAL -'.$periode.' SECOND)
                UNION
                SELECT count(*) activitat FROM actualitzacio WHERE added_on >= DATE_ADD(NOW(), INTERVAL -'.$periode.' SECOND)) activitats');
        return $q->row();
    }
    
    protected function _create_timestamp($data)
    {
       $data['added_on'] = $data['updated_on'] = date('Y-m-d H:i:s');
       $data['updated_by'] = $this->session->userdata('usuari')->id_usuari;
       return $data;
    }

    protected function _update_timestamp($data)
    {
       $data['updated_on'] = date('Y-m-d H:i:s');
       $data['updated_by'] = $this->session->userdata('usuari')->id_usuari;
       return $data;
    }
}