<?php 
/**
 * Model taula 'usuari' i relacionades
 * Basat en MY_Model (funcionalitat CRUD)
 */

if (!defined('BASEPATH')) exit('Accés denegat');

/**
 * 
 */
class Usuari_model extends MY_Model
{
    public $_table = 'usuari';
    public $primary_key = 'id_usuari';
    public $before_create = array( '_create_timestamp' );
    public $before_update = array( '_update_timestamp' );
    
    public function __construct()
    {
        parent::__construct();
    }

    /**
     * Valida si l'usuari i contrasenya són correctes i actius
     * @param {string} $usuari
     * @param {string} $contrasenya
     * @return {boolean}
     */
    public function auth_usuari($usuari, $contrasenya)
    {
        $res = $this->get_usuari($usuari, TRUE);
        $this->load->library('phpass');
        // Usuari no existeix o inactiu
        if ($res == FALSE || ($res->data_baixa != NULL && strtotime($res->data_baixa)<=strtotime(date("Y-m-d"))))
        {
            return FALSE;
        }
        // Usuari existeix, contrasenya correcte
        elseif ($this->phpass->check($contrasenya, $res->contrasenya))
        {
            unset($res->contrasenya);
            return $res;
        }
        // Usuari existeix, contrasenya incorrecte
        else {
            return TRUE;
        }
    }
    
    /**
     * Retorna les dades d'un usuari de la BD
     * @param {String} $id
     * @param {String} $contrasenya
     * @return {Object}
     */
    public function get_usuari($id, $contrasenya = false)
    {
        $this->db->select('id_usuari, contrasenya, usuari.nom, cognoms, email, foto, departament, departament.nom as departament_nom, data_baixa, added_on, updated_on, rol as rol_id, rol.nom as rol', FALSE);
        $this->db->from('usuari');
        $this->db->join('rol', 'rol.id_rol = usuari.rol');
        $this->db->join('departament', 'departament.id_departament = usuari.departament', 'LEFT');
        $this->db->where(array('id_usuari' => $id));
        $query = $this->db->get();
        if ($query->num_rows() > 0)
        {
            $usuari = $query->row();
            if (!$contrasenya) unset($usuari->contrasenya);
            return $usuari;
        }
        return FALSE;
    }
    
    /**
     * Obté usuaris segons filtres donats ordenats per data creació
     * @param {array} $filtres Array de filtres
     * @param {string} $limit Màxim d'usuaris a retornar
     * @param {string} $inici Punter inici pel qual començar a retornar usuaris
     * @return {array} Usuaris
     */
    public function get_usuaris($filtres = array(), $limit = '0', $inici = '0')
    {
        $this->db->select('usuari.id_usuari, usuari.nom, usuari.cognoms, usuari.email, rol.descripcio rol, usuari.rol rol_id, departament.nom departament,
            (CASE WHEN usuari.foto IS NULL THEN "generica.png" ELSE usuari.foto END) foto,
            data_baixa,
            DATE_FORMAT(usuari.added_on, "%d/%m/%Y %H:%i") data_alta,
            DATE_FORMAT(usuari.updated_on, "%d/%m/%Y %H:%i") data_upd', FALSE);
        $this->db->from('usuari');
        $this->db->join('rol', 'rol.id_rol = usuari.rol');
        $this->db->join('departament', 'departament.id_departament = usuari.departament', 'left');
        if (@$filtres['text'])
        {
            $text = $this->db->escape_like_str($filtres['text']);
            $this->db->where('(usuari.id_usuari LIKE "%'.$text.'%" OR usuari.nom LIKE "%'.$text.'%" OR usuari.cognoms LIKE "%'.$text.'%")');
        }
        if (@$filtres['rol'])
            $this->db->where('rol', $filtres['rol']);
        if (@$filtres['actiu'])
            $this->db->where('data_baixa is null or data_baixa = ""');

        $this->db->order_by('added_on desc, id_usuari asc');
        
        if ($limit!='0' || $inici!='0')
            $this->db->limit($limit, $inici);
        
        $res = $this->db->get();
        return $res->result();
    }
    
    /**
     * Guarda usuari amb les dades enviades per formulari
     * @param {Array} $post Dades enviades per formulari
     * @return {String} ID del nou usuari
     */
    public function insert_usuari($input)
    {
        $this->load->library('phpass');
        $data = array(
            'id_usuari' => $input['id_usuari'],
            'contrasenya' => $this->phpass->hash($input['contrasenya']),
            'nom' => $input['nom'],
            'cognoms' => $input['cognoms'],
            'email' => $input['email'],
            'rol' => $input['rol']
        );
        if ($input['rol'] !== '1') $data['departament'] = $input['departament'];
        return $this->db->insert('usuari', $this->_create_timestamp($data));
    }
    
    /**
     * Actualitza usuari amb les dades enviades per formulari
     * @param {Array} $post Dades enviades per formulari
     * @return {String} ID usuari actualitzat
     */
    public function upd_usuari($input)
    {
        $data = array(
            'nom' => $input['nom'],
            'cognoms' => $input['cognoms'],
            'email' => $input['email'],
            'rol' => $input['rol'],
        );
        if ($input['rol'] !== '1') $data['departament'] = $input['departament'];
        if ($input['data_baixa'])
            $data['data_baixa'] = $input['data_baixa'];
        else
            $data['data_baixa'] = NULL;
        
        $this->db->where('id_usuari', $input['id_usuari']);
        return $this->db->update('usuari', $this->_update_timestamp($data));
    }
    
    /**
     * Retorna identificadors usuaris segons el seu rol i el seu nom
     * per mostrar en un input amb autocompletació.
     * @param {string} $rol Filtrar per aquest rol
     * @return {array} Usuaris
     */
    public function get_usuaris_autocomplete($rol)
    {
        $this->db->select('id_usuari, concat(usuari.nom, " ", usuari.cognoms) nom', FALSE);
        $this->db->from('usuari');
        $this->db->join('rol', 'rol.id_rol = usuari.rol');
        if ($rol === 'gestor')
        {
            $this->db->where('rol.nom', $rol);
            $this->db->or_where('rol.nom', 'admin');
        }
        else
        {
            $this->db->where('rol.nom', $rol);
        }
            
        $query = $this->db->get();
        return $query->result();
    }
    
    /**
     * Retorna gestors d'un departament en concret
     * @param {string} $departament Departament dels gestors
     * @return {array} Gestors del departament
     */
    public function get_gestors_departament($departament = '')
    {
        $this->db->select('id_usuari, concat(usuari.nom, " ", usuari.cognoms) nom', FALSE);
        $this->db->from('usuari');
        $this->db->where_in('rol', array(2,3));
        if ($departament !== '')
            $this->db->where('departament', $departament);
        $multi_dim = $this->db->get()->result();
        $gestors = array();
        $gestors[NULL] = '';
        foreach ($multi_dim as $gestor)
        {
            $gestors[$gestor->id_usuari] = $gestor->nom;
        }
        return $gestors;
    }
    
    /**
     * Retorna tots els rols per mostrar-los a un select
     * @return {array} Rols
     */
    public function get_rols()
    {
        $multi_dim = $this->db->get('rol')->result();
        $rols = array();
        $rols[NULL] = '';
        foreach ($multi_dim as $rol)
        {
            $rols[$rol->id_rol] = $rol->descripcio;
        }
        return $rols;
    }
    
    protected function _create_timestamp($data) {
       $data['added_on'] = $data['updated_on'] = date('Y-m-d H:i:s');
       $data['updated_by'] = $this->session->userdata('usuari')->id_usuari;
       return $data;
    }

    protected function _update_timestamp($data) {
       $data['updated_on'] = date('Y-m-d H:i:s');
       $data['updated_by'] = $this->session->userdata('usuari')->id_usuari;
       return $data;
    }
}

/* Fi del fitxer usuari_model.php */
/* Ubicació: ./application/models/usuari_model.php */